from django.utils.translation import gettext_lazy as _

from violation.behaviours.rule import BaseRuleModelMixin as RuleModelMixin


class Rule(RuleModelMixin):
    class Meta:
        app_label = 'violation'
        verbose_name = _('Rule')
        verbose_name_plural = _('Rules')



from django.db import models
from django.contrib.contenttypes.models import ContentType


class ViolationManager(models.Manager):
    def get_violations_for(self, obj, **kwargs):
        return self.filter(
            content_type=ContentType.objects.get_for_model(obj),
            object_id=obj.id,
            **kwargs
        )



class BaseViolationViewMixin:

    """
    This viewmixin supplies every get_form_kwargs needed for initialization
    Override methods here for further customization
    """

    def get_form_kwargs(self):
        """Add request and object keywords to BaseViolationForm()"""
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.add_request()
        kwargs['object'] = self.add_object()
        return kwargs

    def add_request(self):
        """:return: request"""
        return self.request

    def add_object(self, obj=None):
        "Object to be use in form"
        """
        ::param obj
        :return obj
        """
        return obj if obj is not None else self.get_object()



from django.views.generic.edit import CreateView

from violation.forms.violation import ViolationForm
from violation.viewmixins.violation import BaseViolationViewMixin


class BaseViolationView(BaseViolationViewMixin, CreateView):
    """
    To use this view, you must defined a form_class
    that inherits from either BaseViolationForm or ViolationForm
    """



from django.utils.translation import gettext_lazy as _

from violation.behaviours.violation import ViolationRuleModelMixin as ViolationModelMixin


class Violation(ViolationModelMixin):
    class Meta:
        app_label = 'violation'
        verbose_name = _('violation')
        verbose_name_plural = _('violations')



from django.contrib import admin

from violation.models import Rule, Violation


class RuleViolationInline(admin.StackedInline):
    model = Rule.violations.through


@admin.register(Rule)
class RuleAdmin(admin.ModelAdmin):
    radio_fields = {'category': admin.HORIZONTAL}
    list_display = ['name', 'category', 'short_rule', 'added_by']
    list_filter = ['category', 'added_by']
    search_fields = ['name', 'category', 'description']
    ordering = ['name', 'category', 'added_by']
    save_on_top = True

    def short_rule(self, obj):
        return obj.description[:20]

    short_rule.short_description = 'Description'


@admin.register(Violation)
class ViolationAdmin(admin.ModelAdmin):
    readonly_fields = ['violator', 'created', 'modified']
    list_display_links = ['id', 'rules_violated']
    radio_fields = {'status': admin.HORIZONTAL}
    list_display = [
        'id',
        'reported_by',
        'violator',
        'item_of_violation_category',
        'item_of_violation',
        'object_id',
        'rules_violated',
        'date_reported',
        'status',
        'is_violated',
        'last_modified'
    ]
    ordering = [
        'reported_by',
        'violator',
        'is_violated',
        'status'
    ]
    list_filter = [
        'reported_by',
        'violator',
        'status',
        'is_violated'
    ]
    list_editable = ['is_violated', 'status']
    date_hierarchy = 'created'
    filter_horizontal = ['rules']
    list_per_page = 20

    def item_of_violation_category(self, obj):
        return obj.content_type

    item_of_violation_category.short_description = 'Item category'

    def item_of_violation(self, obj):
        return obj.content_object

    item_of_violation.short_description = 'Item of violation'

    def date_reported(self, obj):
        return obj.created

    date_reported.short_description = 'Date reported'

    def last_modified(self, obj):
        return obj.modified

    date_reported.short_description = 'Last action'

    def rules_violated(self, obj):
        rules = obj.rules.all()
        return ('%s' % ', '.join([rule.name for rule in rules]))

    short_description = 'Violated rules'

    def get_readonly_fields(self, request, obj=None):
        if obj is not None:
            return self.list_display
        return []



from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _

User = get_user_model()


class BaseViolationModelMixin(models.Model):

    VIOLATION_STATUS_PENDING = 0 # whether violation is awaiting approval from admin
    VIOLATION_STATUS_ACCEPTED = 1 # whether violation has been accepted by admin
    VIOLATION_STATUS_REJECTED = 2 # whether violation was rejected by admin

    VIOLATION_STATUS_CHOICES = (
        (VIOLATION_STATUS_PENDING, _('Pending')),
        (VIOLATION_STATUS_ACCEPTED, _('Accepted')),
        (VIOLATION_STATUS_REJECTED, _('Rejected')),
    )

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    content_object = GenericForeignKey()
    object_id = models.PositiveIntegerField()

    reported_by = models.ForeignKey(
        User,
        on_delete=models.SET_NULL,
        related_name='violations_reported',
        null=True
    )
    violator = models.ForeignKey(
        User,
        on_delete=models.SET_NULL,
        related_name='violations_broken',
        null=True,
        blank=True
    )
    status = models.PositiveSmallIntegerField(
        choices=VIOLATION_STATUS_CHOICES,
        default=VIOLATION_STATUS_PENDING
    )
    is_violated = models.BooleanField(null=True)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    def save(self, *args, **kwargs):
        from violation.signals import report_handler
        super().save(*args, **kwargs)
        report_handler.send(
            sender=self.__class__,
            violation=self,
            violator=self.content_object
        )

    def __str__(self):
        return f'{self.reported_by} just reported {self.content_object} for violation'

    class Meta:
        abstract = True
        constraints = [
            models.UniqueConstraint(
                fields=['content_type', 'object_id', 'reported_by'],
                name='unique_violation_constraint'
            )
        ]


class ViolationRuleModelMixin(BaseViolationModelMixin):
    rules = models.ManyToManyField('violation.Rule', related_name='violations')

    class Meta:
        abstract = True



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class ViolationConfig(AppConfig):
    name = 'violation'
    verbose_name = _('Violation')

    def ready(self):
        from . import signals



# Generated by Django 3.1 on 2020-08-31 20:11

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Rule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, choices=[('post', 'Post'), ('forum', 'Forum'), ('thread', 'Thread'), ('user', 'User'), ('general', 'General')], max_length=10, null=True, verbose_name='category')),
                ('name', models.CharField(max_length=255, verbose_name='name')),
                ('description', models.TextField(verbose_name='description')),
                ('added_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Added by')),
            ],
            options={
                'verbose_name': 'Rule',
                'verbose_name_plural': 'Rules',
            },
        ),
        migrations.CreateModel(
            name='Violation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('status', models.PositiveSmallIntegerField(choices=[(0, 'Pending'), (1, 'Accepted'), (2, 'Rejected')], default=0)),
                ('is_violated', models.BooleanField(null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('reported_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='violations_reported', to=settings.AUTH_USER_MODEL)),
                ('rules', models.ManyToManyField(related_name='violations', to='violation.Rule')),
                ('violator', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='violations_broken', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'violation',
                'verbose_name_plural': 'violations',
            },
        ),
    ]



from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import gettext_lazy as _

from ..managers.rule import RuleManager

User = get_user_model()


class BaseRuleModelMixin(models.Model):
    CATEGORY_USER = 'user'
    CATEGORY_FORUM = 'forum'
    CATEGORY_POST = 'post'
    CATEGORY_THREAD = 'thread'
    CATEGORY_GENERAL = 'general'

    CATEGORY_CHOICES = [
        (CATEGORY_POST, _('Post')),
        (CATEGORY_FORUM, _('Forum')),
        (CATEGORY_THREAD, _('Thread')),
        (CATEGORY_USER, _('User')),
        (CATEGORY_GENERAL, _('General')),
    ]

    category = models.CharField(
        verbose_name=_('category'),
        max_length=10,
        choices=CATEGORY_CHOICES,
        blank=True,
        null=True
    )
    name = models.CharField(verbose_name=_('name'), max_length=255)
    description = models.TextField(verbose_name=_('description'))
    added_by = models.ForeignKey(
        User,
        verbose_name=_('Added by'),
        on_delete=models.DO_NOTHING,
        related_name='+',
        null=True,
        blank=True
    )
    objects = RuleManager()

    def __str__(self):
        return f'{self.name} - {self.get_category_display()}'

    class Meta:
        abstract = True
        ordering = ['name']



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.db import NotSupportedError
from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _

from violation.fields.rule import RulesModelMultipleChoiceField
from violation.models import Rule, Violation


class BaseViolationForm(forms.ModelForm):
    """ This is the base form for further customization """
    categories = None
    queryset = None
    dont_repeat_rules = False

    """
        This is the base form to display list of rules in a category
        1. ::categories: rules in specified categories to display
        2. ::queryset: rules queryset to display.
        3. ::dont_repeat_rules: allow whether to show already violated rule by an object
        or not. If set to False, rules already violated by an item will still be made available
        for other users to select.
    """

    rules = RulesModelMultipleChoiceField(queryset=None)

    class Meta:
        model = Violation
        fields = ['rules']

    def __init__(self, *args, **kwargs):
        self.object = kwargs.pop('object')
        self.request = kwargs.pop('request')
        self.rules_queryset = self.get_queryset()
        if self.dont_repeat_rules:
            self.rules_queryset = self.unique_rules_queryset()
        super().__init__(*args, **kwargs)
        self.fields['rules'].widget = forms.CheckboxSelectMultiple()
        self.fields['rules'] = RulesModelMultipleChoiceField(
            queryset=self.rules_queryset,
            widget=forms.CheckboxSelectMultiple
        )

    def get_queryset(self):
        if self.categories is None:
            if self.queryset is not None:
                queryset = self.queryset
                if isinstance(self.queryset, QuerySet):
                    queryset = queryset.all()
                else:
                    raise NotSupportedError('%s is actually not a QuerySet' % self.queryset)
            else:
                raise ImproperlyConfigured('self.categories or self.queryset is needed to load this form')
        else:
            queryset = Rule.objects.filter(
                category__in=[category for category in self.categories]
            )
        return queryset

    def clean_rules(self):
        """
        A validation check which ensures a user/item doesnt
        report an item more than ones
        """
        cleaned_rules = self.cleaned_data['rules']
        check_spam_violations = self.check_spam_rules(cleaned_rules, self.request.user)
        if check_spam_violations:
            raise forms.ValidationError(
                _('A rule cannot be reported twice for violation'),
                code='multiple_violations'
            )
        return cleaned_rules

    """
    This method is called to ensure a report action is not repeated twice.
    """

    def check_spam_rules(self, rules, user):
        violations = Violation.objects.filter(
            content_type=ContentType.objects.get_for_model(self.object),
            object_id=self.object.id,
            reported_by=user,
            rules__in=rules
        )
        return violations.exists()

    def unique_rules_queryset(self):
        """This method is called only when self.dont_repeat_rules is set."""
        qs = self.get_queryset()
        return qs.exclude(
            violations__isnull=False,
            violations__status__in=[
                Violation.VIOLATION_STATUS_PENDING,
                Violation.VIOLATION_STATUS_ACCEPTED
            ],
            violations__content_type=ContentType.objects.get_for_model(self.object),
            violations__object_id=self.object.id,
            violations__reported_by=self.request.user,
            violations__is_violated__isnull=True
        )

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.reported_by = self.request.user
        instance.content_type = ContentType.objects.get_for_model(self.object)
        instance.object_id = self.object.id
        if commit:
            instance.save()
        return instance


class ViolationForm(BaseViolationForm):
    """
    This is just a sample form.
    Free to inherit it if you just want simple violation rules display
    """
    categories = ['general']